import lmfit
import numpy as np


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E_S", value=3e3, min=0)
    params.add("E_L", value=20, min=0, max=1000)
    params.add("R", value=10e-6, min=0, vary=False)
    params.add("nu_S", value=.3, min=0, max=0.5, vary=False)
    params.add("nu_L", value=.3, min=0, max=0.5, vary=False)
    params.add("t", value=0.1e-6, min=1e-12)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def power_layer_clifford_2009(delta, E_S, E_L, R, nu_S, nu_L, t,
                              contact_point=0, baseline=0):
    r"""Power law model including one elastic layer covering the sample

    .. math::

        F &= \frac{4}{3}
            E^*
            \sqrt{R}
            \delta^{3/2}

        E^* &= E_\mathrm{L}
              + (E_\mathrm{S} - E_\mathrm{L})
              \frac{P \xi^n}{1 + P \xi^n}

        \xi &= \frac{\sqrt{R\delta}}{t}
               \left(
               \frac{E_\mathrm{L}}{E_\mathrm{S}}
               \right)^m
               \frac{1 - B_\mathrm{S}\nu_\mathrm{S}^2}
               {1 - B_\mathrm{L}\nu_\mathrm{L}^2}

    with the fixed constants determined from the finite element fits
    :math:`P=2.25`, :math:`n=1.5`, :math:`m=2/3`,
    :math:`B_\mathrm{S}=0.22`, and :math:`B_\mathrm{L}=1.92`.


    Parameters
    ----------
    delta: 1d ndarray
        Indentation [m]
    E_S: float
        Young's modulus of the sample [N/m²]
    E_L: float
        Young's modulus of the layer [N/m²]
    R: float
        Tip radius [m]
    nu_S: float
        Poisson's ratio of the sample
    nu_L: float
        Poisson's ratio of the layer
    t: float
        Thickness of the layer covering the sample [m]
    contact_point: float
        Indentation offset [m]
    baseline: float
        Force offset [N]

    Returns
    -------
    F: float
        Force [N]

    Notes
    -----
    The power law model was derived by fitting finite element
    model simulations for materials in the ~100 GPa range
    and a layer thickness in the range of ~1 µm.

    Neither for :math:`t \rightarrow 0`
    nor for :math:`E_\mathrm{L} \rightarrow E_\mathrm{S}`
    will the power law model converge to the paraboloidal hertz model.

    References
    ----------
    Clifford (2009) :cite:`Clifford2009` (equations 9 and 10)
    """
    # Note that this code is optimized for readability, not for speed.
    # roots of delta
    root = contact_point - delta
    pos = root > 0
    dr12 = np.zeros_like(delta)
    dr12[pos] = np.sqrt(root[pos])
    dr32 = np.zeros_like(delta)
    dr32[pos] = root[pos]**(3/2)
    # constants
    P = 2.25
    n = 1.5
    m = 2/3
    B_S = 0.22
    B_L = 1.92
    # inner term (equation 9)
    a = np.sqrt(R) * dr12
    xi = (a / t
          * (E_L/E_S)**m
          * (1 - B_S * nu_S**2) / (1 - B_L * nu_L**2)
          )
    # outer term for emodulus (equation 10)
    E = E_L + (E_S - E_L) * (P * xi**n) / (1 + P * xi**n)
    # original "hertz" first term
    hertz = 4/3 * E * np.sqrt(R) * dr32
    return hertz + baseline


model_doc = power_layer_clifford_2009.__doc__
model_func = power_layer_clifford_2009
model_key = "power_layer_clifford_2009"
model_name = "elastic layer power law (Clifford 2009)"
parameter_keys = ["E_S", "E_L", "R", "nu_S", "nu_L", "t",
                  "contact_point", "baseline"]
parameter_names = ["Sample Young's Modulus", "Layer Young's Modulus",
                   "Tip Radius", "Sample Poisson's Ratio",
                   "Layer Poisson's Ratio", "Layer Thickness",
                   "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "Pa", "m", "", "", "m", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# nanite documentation build configuration file, created by
# sphinx-quickstart on Tue Sep 26 17:55:31 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import pathlib
import matplotlib
import mock
import os.path as op
import sys
import tomllib
import nanite

# parent directory
pdir = op.dirname(op.dirname(op.abspath(__file__)))
# include extensions
sys.path.append(op.abspath('extensions'))

matplotlib.use("agg")

for mod in ["tkinter",
            "matplotlib.backends.backend_tkagg"]:
    sys.modules[mod] = mock.Mock()

# http://www.sphinx-doc.org/en/stable/ext/autodoc.html#confval-autodoc_member_order
# Order class attributes and functions in separate blocks
autodoc_member_order = 'groupwise'
autoclass_content = 'both'

# include source of matplotlib plots
plot_include_source = True

# Display link to GitHub repo instead of doc on rtfd
rst_prolog = """
:github_url: https://github.com/AFM-Analysis/nanite
"""

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.intersphinx',
              'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon',
              'sphinxcontrib.bibtex',
              'matplotlib.sphinxext.plot_directive',
              'IPython.sphinxext.ipython_directive',
              'IPython.sphinxext.ipython_console_highlighting',
              'simple_argparse',
              'fancy_include',
              'nanite_models',
              'nanite_preprocs',
              'github_changelog',
              ]

# specify bibtex files (required for sphinxcontrib.bibtex>=2.0)
bibtex_bibfiles = ['nanite.bib']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'nanite'
github_project = 'AFM-Analysis/' + project

with (pathlib.Path(__file__).parent.parent / "pyproject.toml").open("rb") as f:
    data = tomllib.load(f)
authors = [a["name"] for a in data["project"]["authors"]]
author = ", ".join(authors)
copyright = '2015, ' + author

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
#
# The full version, including alpha/beta/rc tags.
# This gets 'version'
release = nanite.__version__
# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

numfig = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['data']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'nanitedoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'nanite.tex', 'nanite Documentation',
     author.replace(', ', '\\and '), 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'nanite', 'nanite Documentation',
     authors, 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'nanite', 'nanite Documentation',
     author, 'nanite', 'Loading, fitting, and rating AFM nanoindentation data',
     'Scientific'),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ('https://docs.python.org/', None),
    "numpy": ('http://docs.scipy.org/doc/numpy', None),
    "scipy": ('https://docs.scipy.org/doc/scipy/reference/', None),
    "sklearn": ('https://scikit-learn.org/stable/', None),
    "afmformats": ('https://afmformats.readthedocs.io/en/stable', None),
                       }



"""nanite preprocessing summary

Usage
-----
Directives:

Table of all preprocessors available in nanite

   .. nanite_preproc_table::

Table of all POC methods in nanite

   .. nanite_preproc_poc_table::

"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from nanite import poc, preproc


class Base(Directive):
    required_arguments = 0
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class PreprocTable(Base):
    def generate_rst(self):
        rst = []

        rst.append(".. csv-table::")
        rst.append("    :header: preprocessor key, description, details")
        rst.append("    :delim: tab")
        rst.append("")

        for pp in preproc.PREPROCESSORS:
            ref = f"nanite.preproc.preproc_{pp.identifier}"
            details = ":func:`code reference <{}>`".format(ref)
            rst.append(f"    {pp.identifier}\t {pp.name}\t {details}")

        rst.append("")

        return rst


class PreprocPOCTable(Base):
    def generate_rst(self):
        rst = []

        pocs = poc.POC_METHODS

        rst.append(".. csv-table::")
        rst.append("    :header: POC method, description, details")
        rst.append("    :delim: tab")
        rst.append("")

        for pp in pocs:
            ref = f"nanite.poc.{pp.__name__}"
            details = ":func:`code reference <{}>`".format(ref)
            method = pp.identifier
            name = pp.name
            rst.append(f"    {method}\t {name}\t {details}")

        rst.append("")

        return rst


def setup(app):
    app.add_directive('nanite_preproc_table', PreprocTable)
    app.add_directive('nanite_preproc_poc_table', PreprocPOCTable)
    return {'version': '0.2'}   # identifies the version of our extension



"""nanite models summary

Usage
-----
Directives:

Table of all models available in nanite, including a texed model function

   .. nanite_model_table::


Documents all available models (for code reference)

   .. nanite_model_doc::

"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from nanite import model


class Base(Directive):
    required_arguments = 0
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class ModelDoc(Base):
    def generate_rst(self):
        rst = []

        keys = sorted(model.models_available.keys())

        for kk in keys:
            mod = model.models_available[kk]
            rst.append("")
            rst.append(".. _sec_ref_model_{}:".format(kk))
            rst.append("")
            rst.append(mod.model_name)
            rst.append("~" * len(mod.model_name))

            rst.append(".. csv-table::")
            rst.append("    :delim: tab")
            rst.append("")
            rst.append("    model key\t {}".format(mod.model_key))
            rst.append("    model name\t {}".format(mod.model_name))
            rst.append("    model location\t {}".format(mod.module.__name__))
            rst.append("")

            rst.append(".. automodule:: {}".format(mod.module.__name__))
            rst.append("    :members:")
            rst.append("    :undoc-members:")
            rst.append("    :exclude-members: model, residual, "
                       + "get_parameter_defaults, model_func")
            rst.append("")

        return rst


class ModelTable(Base):
    def generate_rst(self):
        rst = []

        rst.append(".. csv-table::")
        rst.append("    :header: model key, description, details")
        rst.append("    :delim: tab")
        rst.append("")

        keys = sorted(model.models_available.keys())

        for kk in keys:
            mod = model.models_available[kk]
            ref = "sec_ref_model_{}".format(kk)
            details = ":ref:`code reference <{}>`".format(ref)
            rst.append("    {}\t {}\t {}".format(kk, mod.model_name, details))

        rst.append("")

        return rst


def setup(app):
    app.add_directive('nanite_model_table', ModelTable)
    app.add_directive('nanite_model_doc', ModelDoc)
    return {'version': '0.1'}   # identifies the version of our extension



"""Test preprocessing"""
import pathlib

import numpy as np
import pytest

from nanite import IndentationGroup
from nanite import preproc


data_path = pathlib.Path(__file__).resolve().parent / "data"
bad_files = list(data_path.glob("bad*"))


def test_autosort():
    unsorted = ["correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position"]
    expected = ["correct_force_offset",
                "compute_tip_position",
                "correct_tip_offset"]
    actual = preproc.autosort(unsorted)
    assert expected == actual


def test_autosort2():
    unsorted = ["correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position",
                ]
    expected = ["compute_tip_position",
                "correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                ]
    actual = preproc.autosort(unsorted)
    assert (actual.index("correct_split_approach_retract")
            > actual.index("compute_tip_position"))
    assert (actual.index("correct_tip_offset")
            > actual.index("compute_tip_position"))
    assert expected == actual


def test_autosort3():
    unsorted = ["smooth_height",
                "correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position",
                ]
    expected = ["compute_tip_position",
                "correct_split_approach_retract",
                "smooth_height",
                "correct_force_offset",
                "correct_tip_offset",
                ]
    actual = preproc.autosort(unsorted)
    assert expected == actual
    assert (actual.index("correct_split_approach_retract")
            > actual.index("compute_tip_position"))
    assert (actual.index("correct_tip_offset")
            > actual.index("compute_tip_position"))
    assert (actual.index("smooth_height")
            > actual.index("correct_split_approach_retract"))


def test_check_order():
    with pytest.raises(ValueError, match="Wrong optional step order"):
        preproc.check_order([
            "smooth_height",
            "correct_split_approach_retract"])


def test_get_steps_required():
    req_act = preproc.get_steps_required("correct_tip_offset")
    req_exp = ["compute_tip_position"]
    assert req_act == req_exp


def test_preproc_correct_force_slope_bad_monotonic_data_issue_25():
    fd = IndentationGroup(
        data_path
        / "fmt-jpk-fd_single_tilted-baseline-drift-"
          "mitotic_2021-01-29.jpk-force")[0]
    # put a custom, bad tip position (this requires some hacking)
    s = len(fd["force"])
    raw_data = {}
    for col in fd.columns_innate:
        raw_data[col] = fd[col]
    tippos = np.linspace(0, 5e-9, s)
    tippos = np.roll(tippos, s//2)
    tippos[:s//2] = 0
    raw_data["tip position"] = tippos
    fd._raw_data = raw_data
    # sanity check
    assert np.all(fd["tip position"] == tippos)
    assert "tip position" in fd.columns_innate
    # This caused "TypeError: expected non-empty vector for x" in
    # np.polynomial in nanite 3.7.3.
    fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset", "correct_force_slope",
         "correct_force_offset"],
        options={
            "correct_tip_offset": {"method": "deviation_from_baseline"},
            "correct_force_slope": {"region": "approach",
                                    "strategy": "drift"},
        },
        ret_details=True)


def test_preproc_correct_force_slope_drift_approach():
    fd = IndentationGroup(
        data_path
        / "fmt-jpk-fd_single_tilted-baseline-drift-"
          "mitotic_2021-01-29.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset", "correct_force_slope",
         "correct_force_offset"],
        options={
            "correct_tip_offset": {"method": "fit_line_polynomial"},
            "correct_force_slope": {"region": "approach",
                                    "strategy": "drift"},
        },
        ret_details=True)
    slopedet = details["correct_force_slope"]
    for key in ["plot slope data",
                "plot slope fit",
                "norm"]:
        assert key in slopedet
    # Sanity check for size of baseline (determined by POC via line+poly)
    assert len(slopedet["plot slope data"][0]) == 15602
    # Check for flatness of baseline.
    assert np.ptp(fd["force"][:15000]) < .094e-9, "ptp less than 0.1 nN"
    bl0 = np.mean(fd["force"][:100])
    bl1 = np.mean(fd["force"][10000:10100])
    assert np.allclose(bl0, bl1,
                       atol=1e-11,
                       rtol=0)
    assert np.abs(bl1 - bl0) < 0.0085e-9
    # Now check whether the maximum force is lower than normal.
    assert np.allclose(np.max(fd["force"]), 1.6688553846662955e-09,
                       atol=0)
    assert np.allclose(np.min(fd["force"]), -5.935613094149927e-10,
                       atol=0)
    # Compare it to the correction obtained with just "baseline".
    # For this dataset, the drift is resulting in a continuous increase
    # in the force. Thus, if we compare the "approach" correction with
    # just the "baseline" correction, we should arrive at a lower number
    # for the "approach" correction.
    assert np.max(fd["force"]) < 1.6903757572616587e-09


def test_preproc_correct_force_slope_drift_full():
    fd = IndentationGroup(
        data_path
        / "fmt-jpk-fd_single_tilted-baseline-drift-"
          "mitotic_2021-01-29.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset", "correct_force_slope"],
        options={
            "correct_tip_offset": {"method": "fit_line_polynomial"},
            "correct_force_slope": {"region": "all",
                                    "strategy": "drift"},
        },
        ret_details=True)
    slopedet = details["correct_force_slope"]
    for key in ["plot slope data",
                "plot slope fit",
                "norm"]:
        assert key in slopedet
    # Sanity check for size of baseline (determined by POC via line+poly)
    assert len(slopedet["plot slope data"][0]) == 15602
    # Check for flatness of baseline.
    assert np.ptp(fd["force"][:15000]) < .094e-9, "ptp less than 0.1 nN"
    bl0 = np.mean(fd["force"][:100])
    bl1 = np.mean(fd["force"][10000:10100])
    assert np.allclose(bl0, bl1,
                       atol=0,
                       rtol=.01)
    assert np.abs(bl1 - bl0) < 0.0085e-9
    # Now check for flatness of retract tail
    rt0 = np.mean(fd["force"][-100:])
    rt1 = np.mean(fd["force"][-10100:-10000])
    assert rt1 < rt0, "There is still a little drift left in the retract tail"
    # The following test would fail if we set "strategy" to "shift" or if
    # we set region to "baseline" or "approach".
    assert np.allclose(rt0, rt1,
                       atol=5e-11, rtol=0)


def test_preproc_correct_force_slope_shift():
    fd = IndentationGroup(
        data_path
        # Note: this dataset actually resembles a drift over time, not
        # a shift. But we are only looking at the baseline, so it's OK.
        / "fmt-jpk-fd_single_tilted-baseline-drift-"
          "mitotic_2021-01-29.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset", "correct_force_slope"],
        options={
            "correct_tip_offset": {"method": "fit_line_polynomial"},
            "correct_force_slope": {"region": "baseline",
                                    "strategy": "shift"},
        },
        ret_details=True)
    slopedet = details["correct_force_slope"]
    for key in ["plot slope data",
                "plot slope fit",
                "norm"]:
        assert key in slopedet
    # Sanity check for size of baseline (determined by POC via line+poly)
    assert len(slopedet["plot slope data"][0]) == 15602
    # Check for flatness of baseline.
    assert np.ptp(fd["force"][:15000]) < .095e-9, "ptp less than 0.1 nN"
    bl0 = np.mean(fd["force"][:100])
    bl1 = np.mean(fd["force"][10000:10100])
    assert np.allclose(bl0, bl1,
                       atol=0,
                       rtol=.01)
    assert np.abs(bl1 - bl0) < 0.009e-9


def test_preproc_correct_force_slope_shift_control():
    """Same test as above, but without slope correction"""
    fd = IndentationGroup(
        data_path
        # Note: this dataset actually resembles a drift over time, not
        # a shift. But we are only looking at the baseline, so it's OK.
        / "fmt-jpk-fd_single_tilted-baseline-drift-"
          "mitotic_2021-01-29.jpk-force")[0]
    fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset"],
        options={
            "correct_tip_offset": {"method": "fit_line_polynomial"},
        })
    # Make sure the baseline is not as flat as in the test above.
    assert np.ptp(fd["force"][:15000]) > .35e-9, "larger ptp with tilt"
    bl0 = np.mean(fd["force"][:100])
    bl1 = np.mean(fd["force"][10000:10100])
    assert not np.allclose(bl0, bl1,
                           atol=0,
                           rtol=.01)
    assert bl1 > bl0
    assert bl1 - bl0 > 0.19e-9


def test_preproc_correct_force_slope_shift_full():
    fd = IndentationGroup(
        data_path
        / "fmt-jpk-fd_single_tilted-baseline-shift-"
          "adyp_2023-06-26.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset", "correct_force_slope"],
        options={
            "correct_tip_offset": {"method": "fit_line_polynomial"},
            "correct_force_slope": {"region": "all",
                                    "strategy": "shift"},
        },
        ret_details=True)
    slopedet = details["correct_force_slope"]
    for key in ["plot slope data",
                "plot slope fit",
                "norm"]:
        assert key in slopedet
    # Sanity check for size of baseline (determined by POC via line+poly)
    assert len(slopedet["plot slope data"][0]) == 3201
    # Check for flatness of baseline.
    assert np.ptp(fd["force"][:3000]) < .04e-9, "ptp less than 0.04 nN"
    bl0 = np.mean(fd["force"][:50])
    bl1 = np.mean(fd["force"][3000:3050])
    assert np.allclose(bl0, bl1,
                       atol=0.004e-9,
                       rtol=0)
    assert np.abs(bl1 - bl0) < 0.0085e-9
    # Now check for flatness of retract tail
    rt0 = np.mean(fd["force"][-50:])
    rt1 = np.mean(fd["force"][-1050:-1000])
    # The following test would fail if we set "strategy" to "drift" or if
    # we set region to "baseline" or "approach".
    assert np.allclose(rt0, rt1,
                       atol=2e-11, rtol=0)


def test_preproc_correct_split_approach_retract():
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]

    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"])
    assert fd.appr["segment"].size == 2000
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset",
                            "correct_split_approach_retract"])
    assert fd.appr["segment"].size == 2006


@pytest.mark.filterwarnings('ignore::nanite.preproc.CannotSplitWarning',
                            'ignore::UserWarning')
def test_process_bad():
    for bf in bad_files:
        ds = IndentationGroup(bf)
        for idnt in ds:
            # Walk through the standard analysis pipeline without
            # throwing any exceptions.
            idnt.apply_preprocessing(["compute_tip_position",
                                      "correct_force_offset",
                                      "correct_tip_offset",
                                      "correct_split_approach_retract"])
            idnt.fit_model()


@pytest.mark.filterwarnings(
    'ignore::nanite.smooth.DoubledSmoothingWindowWarning:')
def test_smooth():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset"])
    orig = np.array(idnt["tip position"], copy=True)
    # now apply smoothing filter
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset",
                              "smooth_height"])
    new = np.array(idnt["tip position"], copy=True)
    assert not np.all(orig == new)


def test_unknown_method():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(KeyError, match="unknown_method"):
        idnt.apply_preprocessing(["compute_tip_position",
                                  "unknown_method"])


def test_wrong_order():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(ValueError, match="requires the steps"):
        # order matters
        idnt.apply_preprocessing(["correct_tip_offset",
                                  "compute_tip_position"])


def test_wrong_order_2():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(ValueError, match="requires the steps"):
        # order matters
        idnt.apply_preprocessing(["correct_split_approach_retract",
                                  "compute_tip_position"])



import copy
import inspect
import functools
import warnings

import lmfit
import numpy as np

from . import poc
from .smooth import smooth_axis_monotone


#: Available preprocessors
PREPROCESSORS = []


class CannotSplitWarning(UserWarning):
    pass


def apply(apret, identifiers=None, options=None, ret_details=False,
          preproc_names=None):
    """Perform force-distance preprocessing steps

    Parameters
    ----------
    apret: nanite.Indentation
        The afm data to preprocess
    identifiers: list
        A list of preprocessing identifiers that will be
        applied (in the order given).
    options: dict of dict
        Preprocessing options for each identifier
    ret_details:
        Return preprocessing details dictionary
    preproc_names: list
        Deprecated - use `identifiers` instead
    """
    if preproc_names is not None:
        identifiers = preproc_names
        warnings.warn(
            "Please use 'identifiers' instead of 'preproc_names'!",
            DeprecationWarning)
    details = {}
    # Reset all user-defined data of the dataset, because we
    # probably edited "tip position", "force", etc. If this
    # is too slow in the future, then we would have to remember
    # the previous preprocessing pipeline (such code is currently
    # implemented on the other side in `indent.py` (2021-08-16).
    apret.reset_data()
    for ii, pid in enumerate(identifiers):
        if pid in available():
            meth = get_func(pid)
            req = meth.steps_required
            act = identifiers[:ii]
            if req is not None and ((set(req) & set(act)) != set(req)):
                raise ValueError(f"The preprocessing step '{pid}' requires"
                                 f" the steps {meth.steps_required}!")
            # create a copy of the dictionary (if it exists) so that
            # `ret_details` is not written to it
            kwargs = copy.deepcopy(options.get(pid, {}))
            if "ret_details" in inspect.signature(meth).parameters:
                # only set `ret_details` if method accepts it
                kwargs["ret_details"] = ret_details
            details[pid] = meth(apret, **kwargs)
        else:
            msg = "The preprocessing method '{}' does not exist!"
            raise KeyError(msg.format(pid))
    # only return details if required
    return details if ret_details else None


def autosort(identifiers):
    """Automatically sort preprocessing identifiers

    This takes into account `steps_required` and `steps_optional`.
    """
    sorted_identifiers = copy.copy(identifiers)
    for pid in identifiers:
        meth = get_func(pid)
        steps_precursor = []
        if meth.steps_required is not None:
            steps_precursor += meth.steps_required
        if meth.steps_optional is not None:
            for ostep in meth.steps_optional:
                if ostep in identifiers:
                    steps_precursor.append(ostep)
        for step in steps_precursor:
            # We have a requirement, check whether it is fulfilled
            cix = sorted_identifiers.index(pid)
            rix = sorted_identifiers.index(step)
            if rix > cix:
                # We pop the wrong requirement and insert it before
                # the current pid.
                sorted_identifiers.remove(step)
                sorted_identifiers.insert(cix, step)

    # Perform a sanity check
    check_order(sorted_identifiers)

    return sorted_identifiers


@functools.lru_cache()
def available():
    """Return list of available preprocessor identifiers"""
    av = [pp.identifier for pp in PREPROCESSORS]
    return autosort(av)


def check_order(identifiers):
    """Check preprocessing steps for correct order"""
    for cix, pid in enumerate(identifiers):
        meth = get_func(pid)
        if meth.steps_required:
            rix = [identifiers.index(r) for r in meth.steps_required]
            if np.any(np.array(rix) > cix):
                raise ValueError(
                    f"Wrong required step order for {pid}: {identifiers}!")
        if meth.steps_optional:
            rio = []
            for rr in meth.steps_optional:
                if rr in identifiers:
                    rio.append(identifiers.index(rr))
            if np.any(np.array(rio) > cix):
                raise ValueError(
                    f"Wrong optional step order for {pid}: {identifiers}!")


def find_turning_point(tip_position, force, contact_point_index):
    """Compute the turning point for a force-indentation curve

    The turning point is defined as the point that is farthest away
    from the contact point in the direction of indentation. This
    implementation normalizes tip position according to their
    minimum and maximum values. This is necessary, because they
    live on different orders of magnitudes/units.
    """
    x = np.copy(tip_position)
    y = np.copy(force)
    idp = contact_point_index

    # Flip and normalize tip position so that maximum is at minimum
    # z-position (set to 1) which coincides with maximum indentation.
    x -= x[idp]
    xmin = x.min()
    if xmin != 0:
        x /= x.min()
    x[x < 0] = 0

    # Flip and normalize force so that maximum force is set to 1.
    y -= np.average(y[:idp])
    y /= y.max()
    y[y < np.std(y[:idp])] = 0

    idturn = np.argmax(x ** 2 + y ** 2)
    return idturn


def get_func(identifier):
    """Return preprocessor function for identifier"""
    for func in PREPROCESSORS:
        if func.identifier == identifier:
            return func
    else:
        raise KeyError(f"Preprocessor '{identifier}' unknown!")


def get_name(identifier):
    """Return preprocessor name for identifier"""
    func = get_func(identifier)
    return func.name


def get_steps_required(identifier):
    """Return requirement identifiers for identifier"""
    func = get_func(identifier)
    return func.steps_required


def preprocessing_step(identifier, name, steps_required=None,
                       steps_optional=None, options=None):
    """Decorator for Indentation preprocessors

    The name and identifier are stored as a property of the wrapped
    function.

    Parameters
    ----------
    identifier: str
        identifier of the preprocessor (e.g. "correct_tip_offset")
    name: str
        human-readble name of the preprocessor
        (e.g. "Estimate contact point")
    steps_required: list of str
        list of preprocessing steps that must be added before this
        step
    steps_optional: list of str
        unlike `steps_required`, these steps do not have to be set,
        but if they are set, they should come before this step
    options: list of dict
        if the preprocessor accepts optional keyword arguments,
        this list yields valid values or dtypes
    """

    def attribute_setter(func):
        """Decorator that sets the necessary attributes

        The outer decorator is used to obtain the attributes.
        This inner decorator returns the actual function that
        wraps the preprocessor.
        """
        func.identifier = identifier
        assert isinstance(name, str)
        func.name = name
        func.options = options
        func.steps_required = steps_required
        func.steps_optional = steps_optional
        PREPROCESSORS.append(func)
        return func

    return attribute_setter


@preprocessing_step(identifier="compute_tip_position",
                    name="tip-sample separation")
def preproc_compute_tip_position(apret):
    """Perform tip-sample separation

    Populate the "tip position" column by adding the force
    normalized by the spring constant to the cantilever
    height ("height (measured)").

    This computation correctly reproduces the column
    "Vertical Tip Position" as it is exported by the
    JPK analysis software with the checked option
    "Use Unsmoothed Height".
    """
    has_hm = "height (measured)" in apret
    has_fo = "force" in apret
    has_sc = "spring constant" in apret.metadata
    if "tip position" in apret.columns_innate:
        # nothing to do
        pass
    elif has_hm and has_fo and has_sc:
        k = apret.metadata["spring constant"]
        force = apret["force"]
        zcant = apret["height (measured)"]
        apret["tip position"] = zcant + force / k
    else:
        missing = []
        if not has_hm:
            missing.append("missing data column 'height (measured)'")
        if not has_fo:
            missing.append("missing data column 'force'")
        if not has_sc:
            missing.append("missing metadata 'spring constant'")
        mt = ", ".join(missing)
        raise ValueError("Cannot compute tip position: {}".format(mt))


@preprocessing_step(identifier="correct_force_offset",
                    name="baseline offset correction",
                    steps_optional=["correct_force_slope"]
                    )
def preproc_correct_force_offset(apret):
    """Correct the force offset with an average baseline value
    """
    idp = poc.compute_poc(force=apret["force"],
                          method="deviation_from_baseline")
    if idp:
        apret["force"] = apret["force"] - np.average(apret["force"][:idp])
    else:
        apret["force"] = apret["force"] - apret["force"][0]


@preprocessing_step(identifier="correct_force_slope",
                    name="baseline slope correction",
                    steps_required=["correct_tip_offset"],
                    options=[
                        {"name": "region",
                         "type": str,
                         "choices": ["baseline", "approach", "all"],
                         "choices_human_readable": [
                             "Correct up until contact point",
                             "Correct the approach part",
                             "Correct the entire dataset"],
                         },
                        {"name": "strategy",
                         "type": str,
                         "choices": ["drift", "shift"],
                         "choices_human_readable": [
                             "Temporal drift correction",
                             "Spatial shift correction"]
                         }
                    ])
def preproc_correct_force_slope(apret, region="baseline", strategy="shift",
                                ret_details=False):
    """Subtract a linear slope from selected parts of the force curve

    Slope correction is a debatable topic in AFM analysis. Many voices
    are of the opinion that this falsifies data and should not be done.
    But when data are scarce, slope correction can help to extract valuable
    information.

    Slope correction uses the baseline (the part of the curve before the POC)
    to determine the slope that should be subtracted from the data. Note
    that for tilted data, you should use a contact point estimate based
    on a line and a polynomial.

    There are three regions for curve correction:

    - baseline: Only the data leading up to the contact point are modified.
      This has the advantage that indentation data are not modified, but the
      baseline can still be used (as weight) in the fitting scheme.
    - appraoch: The slope is subtracted from the entire approach curve.
      This makes sense when you know that the slope affects baseline and
      indentation part of your dataset.
    - all: The entire dataset is corrected. This makes sense if you would
      like to extract information about e.g. viscosity from the area between
      the approach and retract curves of your dataset.

    In addition, there are two strategies available:

    - The "drift" approach assumes that the there is a global drift in the
      dataset (caused e.g. by a thermal drift). Since these things are
      usually of temporal nature, the correction in done over the time axis.
      Use this if the beggining of the approach part and the end of the
      retract part of your dataset "stick out" in opposite directions when
      plotting force over tip position.
    - The "shift" approach assumes that there is a global shift that is a
      function of the distance between sample and cantilever. Use this if
      the beginning of the approach part and the end of the retract part
      align when plotting force over tip position.
    """
    tip_position = apret["tip position"]
    time_position = apret["time"]
    force = apret["force"]

    # Get the current contact point position computed by "correct_tip_offset".
    idp = max(2, np.argmin(np.abs(tip_position)))
    # Determine whether we want to do temporal or spatial correction:
    # Fit a linear slope to the baseline part (all data up until idp)
    mod = lmfit.models.LinearModel()
    if strategy == "shift":
        abscissa = tip_position
    elif strategy == "drift":
        abscissa = time_position
    else:
        raise ValueError(f"Invalid strategy '{strategy}'!")
    pars = mod.guess(force[:idp], x=abscissa[:idp])
    out = mod.fit(force[:idp], pars, x=abscissa[:idp])

    force_edit = np.copy(force)
    # Subtract the linear slope from the region data.
    if region == "baseline":
        # Only subtract the force from data up until the contact point.
        # Make sure that there is no offset/jump by pulling the last
        # element of the best fit array to zero.
        force_edit[:idp] -= out.best_fit - out.best_fit[-1]
    elif region == "approach":
        # Subtract the force from everything that is part of the
        # indentation part.
        idturn = find_turning_point(tip_position=tip_position,
                                    force=force_edit,
                                    contact_point_index=idp)
        idturn = max(2, idturn)
        # Extend the best fit towards the turning point.
        best_fit_approach = mod.eval(out.params, x=abscissa[:idturn])
        force_edit[:idturn] -= best_fit_approach - best_fit_approach[-1]
    elif region == "all":
        # Use the same approach as above, but subtract from the entire
        # curve.
        best_fit_all = mod.eval(out.params, x=abscissa)
        force_edit -= best_fit_all - best_fit_all[idp]
    else:
        raise ValueError(f"Invalid region '{region}'!")

    # Override the force information
    apret["force"] = force_edit

    if ret_details:
        return {
            "plot slope data": [np.arange(idp), force[:idp]],
            "plot slope fit": [np.arange(idp), out.best_fit],
            "norm": "force"}


@preprocessing_step(
    identifier="correct_tip_offset",
    name="contact point estimation",
    steps_required=["compute_tip_position"],
    options=[
        {"name": "method",
         "type": str,
         "choices": [p.identifier for p in poc.POC_METHODS],
         "choices_human_readable": [p.name for p in poc.POC_METHODS]}
    ]
)
def preproc_correct_tip_offset(apret, method="deviation_from_baseline",
                               ret_details=False):
    """Estimate the point of contact

    An estimate of the contact point is subtracted from the
    tip position.
    """
    data = poc.compute_poc(force=apret["force"],
                           method=method,
                           ret_details=ret_details)
    if ret_details:
        cpid, details = data
    else:
        cpid, details = data, None
    apret["tip position"] = (apret["tip position"]
                             - apret["tip position"][cpid])
    return details


@preprocessing_step(identifier="correct_split_approach_retract",
                    name="segment discovery",
                    steps_required=["compute_tip_position"],
                    steps_optional=["correct_force_slope"],
                    )
def preproc_correct_split_approach_retract(apret):
    """Split the approach and retract curves (farthest point method)

    Approach and retract curves are defined by the microscope. When the
    direction of piezo movement is flipped, the force at the sample tip
    is still increasing. This can be either due to a time lag in the AFM
    system or due to a residual force acting on the sample due to the
    bent cantilever.

    To repair this time lag, we append parts of the retract curve to the
    approach curve, such that the curves are split at the minimum height.
    """
    force = apret["force"]

    idp = poc.poc_deviation_from_baseline(force)
    if idp and not np.isnan(idp):
        idturn = find_turning_point(
            tip_position=apret["tip position"],
            force=force,
            contact_point_index=idp,
        )

        segment = np.zeros(len(apret), dtype=np.uint8)
        segment[idturn:] = 1
        apret["segment"] = segment
    else:
        msg = "Cannot correct splitting of approach and retract curve " + \
              "because the contact point position could not be estimated."
        warnings.warn(msg, CannotSplitWarning)


@preprocessing_step(identifier="smooth_height",
                    name="monotonic height data",
                    steps_optional=[
                        # Otherwise we lose the location of the point
                        # of deepest indentation:
                        "correct_split_approach_retract",
                        # Otherwise it might not be applied to
                        # "tip position":
                        "compute_tip_position",
                        # Slope subtraction should happen first:
                        "correct_force_slope"
                        ])
def preproc_smooth_height(apret):
    """Make height data monotonic

    For the columns "height (measured)", "height (piezo), and
    "tip position", this method ensures that the approach and
    retract segments are monotonic.
    """
    orig = ["height (measured)",
            "height (piezo)",
            "tip position"]
    for col in orig:
        if col not in apret:
            continue
        # Apply smoothing
        sm_app = smooth_axis_monotone(apret.appr[col])
        sm_ret = smooth_axis_monotone(apret.retr[col])

        # Replace the column data
        apret.appr[col] = sm_app
        apret.retr[col] = sm_ret


def _deprecate_call(method):
    def wrapper(*args, **kwargs):
        warnings.warn(
            "Using `IndentationPreprocessor` class is deprecated, please use "
            f"the method '{method.__module__}.{method.__name__}' instead!",
            DeprecationWarning)
        return method(*args, **kwargs)
    return wrapper


class IndentationPreprocessor:
    apply = _deprecate_call(apply)
    autosort = _deprecate_call(autosort)
    get_func = _deprecate_call(get_func)
    get_name = _deprecate_call(get_name)
    get_steps_required = _deprecate_call(get_steps_required)
    check_order = _deprecate_call(check_order)
    available = _deprecate_call(available)



r"""Approximating the Hertzian model with a spherical indenter

There is no closed form for the Hertzian model with a spherical indenter.
The force :math:`F` does not directly depend on the indentation depth
:math:`\delta`, but has an indirect dependency via the radius of the circular
contact area between indenter and sample :math:`a` :cite:`Sneddon1965`:

.. math::

    F &= \frac{E}{1-\nu^2} \left( \frac{R^2+a^2}{2} \ln \! \left(
         \frac{R+a}{R-a}\right) -aR  \right) \label{eq:1}\\
    \delta &= \frac{a}{2} \ln
         \! \left(\frac{R+a}{R-a}\right) \label{eq:2}

Here, :math:`E` is the Young's modulus, :math:`R` is the radius of the
indenter, and :math:`\nu` is the Poisson's ratio of the probed material.

Because of this indirect dependency, fitting this model to experimental
data can be time-consuming. Therefore, it is beneficial to approximate this
model with a polynomial function around small values of :math:`\delta/R`
using the Hertz model for a parabolic indenter as a starting point
:cite:`Dobler`:

.. math::

    F = \frac{4}{3} \frac{E}{1-\nu^2} \sqrt{R} \delta^{3/2}
        \left(1
         - \frac{1}{10} \frac{\delta}{R}
         - \frac{1}{840} \left(\frac{\delta}{R}\right)^2
         + \frac{11}{15120} \left(\frac{\delta}{R}\right)^3
         + \frac{1357}{6652800} \left(\frac{\delta}{R}\right)^4
         \right)

This example illustrates the error made with this approach. In nanite, the
model for a spherical indenter has the identifier
:ref:`"sneddon_spher" <sec_ref_model_sneddon_spher>` and the
approximate model has the identifier
:ref:`"sneddon_spher_approx" <sec_ref_model_sneddon_spher_approx>`.

The plot shows the error for the parabolic indenter model
:ref:`"hertz_para" <sec_ref_model_hertz_para>` and for the
approximation to the spherical indenter model.
The maximum indentation depth is set to :math:`R`.
The error made by the approximation of the spherical indenter is more than
four magnitudes lower than the maximum force during indentation.
"""
import matplotlib.pylab as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.lines import Line2D
import matplotlib as mpl
import numpy as np

from nanite.model import models_available

# models
exact = models_available["sneddon_spher"]
approx = models_available["sneddon_spher_approx"]
para = models_available["hertz_para"]
# parameters
params = exact.get_parameter_defaults()
params["E"].value = 1000

# radii
radii = np.linspace(2e-6, 100e-6, 20)

# plot results
plt.figure(figsize=(8, 5))

# overview plot
ax = plt.subplot()
for ii, rad in enumerate(radii):
    params["R"].value = rad
    # indentation range
    x = np.linspace(0, -rad, 300)
    yex = exact.model(params, x)
    yap = approx.model(params, x)
    ypa = para.model(params, x)
    ax.plot(x*1e6, np.abs(yex - yap)/yex.max(),
            color=mpl.cm.get_cmap("viridis")(ii/radii.size),
            zorder=2)
    ax.plot(x*1e6, np.abs(yex - ypa)/yex.max(), ls="--",
            color=mpl.cm.get_cmap("viridis")(ii/radii.size),
            zorder=1)

ax.set_xlabel(r"indentation depth $\delta$ [µm]")
ax.set_ylabel("error in force relative to maximum $F/F_{max}$")
ax.set_yscale("log")
ax.grid()

# legend
custom_lines = [Line2D([0], [0], color="k", ls="--"),
                Line2D([0], [0], color="k", ls="-"),
                ]
ax.legend(custom_lines, ['parabolic indenter',
                         'approximation of spherical indenter'])

divider = make_axes_locatable(ax)
cax = divider.append_axes("right", size="3%", pad=0.05)

norm = mpl.colors.Normalize(vmin=radii[0]*1e6, vmax=radii[-1]*1e6)
mpl.colorbar.ColorbarBase(ax=cax,
                          cmap=mpl.cm.viridis,
                          norm=norm,
                          orientation='vertical',
                          label="indenter radius [µm]"
                          )

plt.tight_layout()
plt.show()



"""Fitting and rating

This example uses a force-distance curve of a zebrafish spinal cord
section to illustrate basic data fitting and rating with nanite.
The dataset is part of a study on spinal cord stiffness in zebrafish
:cite:`Moellmert2019`.
"""
import matplotlib.gridspec as gridspec
import matplotlib.pylab as plt

import nanite

# load the data
group = nanite.load_group("data/zebrafish-head-section-gray-matter.jpk-force")
idnt = group[0]  # this is an instance of `nanite.Indentation`
# apply preprocessing
idnt.apply_preprocessing(["compute_tip_position",
                          "correct_force_offset",
                          "correct_tip_offset"])
# set the fit model ("sneddon_spher_approx" is faster than "sneddon_spher"
# and sufficiently accurate)
idnt.fit_properties["model_key"] = "sneddon_spher_approx"
# get the initial fit parameters
params = idnt.get_initial_fit_parameters()
# set the correct indenter radius
params["R"].value = 18.64e-06
# perform the fit with the edited parameters
idnt.fit_model(params_initial=params)
# obtain the Young's modulus
emod = idnt.fit_properties["params_fitted"]["E"].value
# obtain a rating for the dataset
# (using default regressor and training set)
rate = idnt.rate_quality()

# overview plot
plt.figure(figsize=(8, 5))
gs = gridspec.GridSpec(2, 1, height_ratios=[5, 1])

ax1 = plt.subplot(gs[0])
ax2 = plt.subplot(gs[1])

# only plot the approach part (`1` would be retract)
where_approach = idnt["segment"] == 0

# plot force-distance data (nanite uses SI units)
ax1.plot(idnt["tip position"][where_approach] * 1e6,
         idnt["force"][where_approach] * 1e9,
         label="data")
ax1.plot(idnt["tip position"][where_approach] * 1e6,
         idnt["fit"][where_approach] * 1e9,
         label="fit (spherical indenter)")
ax1.text(.2, 2.05,
         "apparent Young's modulus: {:.0f} Pa\n".format(emod)
         + "rating: {:.1f}".format(rate),
         ha="center")
ax1.legend()
# plot resiudals
ax2.plot(idnt["tip position"][where_approach] * 1e6,
         (idnt["force"] - idnt["fit"])[where_approach] * 1e9)

# update plot parameters
ax1.set_xlim(-4.5, 3)
ax1.set_ylabel("force [pN]")
ax1.grid()
ax2.set_xlim(-4.5, 3)
ax2.set_ylim(-.2, .2)
ax2.set_ylabel("residuals [pN]")
ax2.set_xlabel("tip position [µm]")
ax2.grid()

plt.tight_layout()
plt.show()



"""Test of data set functionalities"""
import pathlib

import lmfit
import numpy as np

from nanite import IndentationGroup
from nanite.model import model_power_layer_clifford_2009 as clifford


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_app_ret():
    ds = IndentationGroup(jpkfile)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset"])
    idp = ar.estimate_contact_point_index()

    aprid = ar["segment"] == 0
    x = ar["tip position"][aprid]
    y = ar["force"][aprid]
    contact_point = x[idp]

    # crop x and y around contact_point
    distcp = x.shape[0]-idp
    x = x[-3*distcp:]
    y = y[-3*distcp:]

    params = clifford.get_parameter_defaults()
    params["t"].vary = False
    params["contact_point"].value = contact_point

    fit_res = lmfit.minimize(clifford.residual, params, args=(x, y, 0))

    # This is only a sanity check to make sure the
    # model is reproducible.
    assert np.allclose(fit_res.params["E_S"].value,
                       42218,
                       rtol=1e-2, atol=0)
    assert np.allclose(fit_res.params["E_L"].value,
                       191,
                       rtol=1e-2, atol=0)
    assert np.allclose(fit_res.params["contact_point"].value,
                       1.805e-5,
                       rtol=1e-2, atol=0)



"""Test usage of user-defined training set"""
import pathlib
import tempfile

import numpy as np

from nanite import IndentationGroup
from nanite.rate import IndentationRater


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"


def setup_training_set(n=300):
    tdir = tempfile.mkdtemp(prefix="test_nanite_rate_ts_")
    tdir = pathlib.Path(tdir)
    np.random.set_state(np.random.RandomState(47).get_state())
    for bb in IndentationRater.get_feature_names(which_type="binary"):
        bvals = np.random.choice([0, 1], size=n, p=[.05, .95])
        np.savetxt(tdir / f"train_{bb}.txt", bvals)
    for cc in IndentationRater.get_feature_names(which_type="continuous"):
        cvals = np.random.random_sample(size=n)
        np.savetxt(tdir / f"train_{cc}.txt", cvals)
    rating = np.random.choice(range(11), size=n)
    np.savetxt(tdir / "train_response.txt", rating)
    return tdir


def test_user_training_set():
    tdir = setup_training_set()
    # load a curve
    idnt = IndentationGroup(jpkfile)[0]
    # fit it
    idnt.fit_model(model_key="sneddon_spher_approx",
                   preprocessing=["compute_tip_position",
                                  "correct_force_offset"])
    r1 = idnt.rate_quality(regressor="Extra Trees", training_set="zef18")
    assert r1 > 9, "sanity check"
    r2 = idnt.rate_quality(regressor="Extra Trees", training_set=tdir)
    assert 4 < r2 < 5, "with the given random state we end up at 4.55"


def test_training_set_impute_nans():
    tdir = setup_training_set()
    # edit one of the training feature data to contain an inf value
    fpath = tdir / "train_feat_con_apr_flatness.txt"
    rpath = tdir / "train_response.txt"

    fdat = np.loadtxt(fpath)
    fdat[10] = 1.1
    fdat[11] = 1.2
    fdat[12] = np.nan
    fdat[13] = np.nan
    np.savetxt(fpath, fdat)

    rdat = np.loadtxt(rpath)
    rdat[rdat == 0] = 1
    rdat[10] = 0
    rdat[11] = 0
    rdat[12] = 0
    rdat[13] = 0
    np.savetxt(rpath, rdat)

    samples, response, names = IndentationRater.load_training_set(
        path=tdir,
        which_type="continuous",
        impute_zero_rated_nan=True,  # This should be the default
        replace_inf=True,  # This should be the default
        ret_names=True,
    )
    idf = names.index("feat_con_apr_flatness")
    data = samples[:, idf]

    assert np.allclose(data[10], 1.1)
    assert np.allclose(data[11], 1.2)
    assert np.allclose(data[12], 1.15)
    assert np.allclose(data[13], 1.15)


def test_training_set_inf_values():
    tdir = setup_training_set()
    # edit one of the training feature data to contain an inf value
    fpath = tdir / "train_feat_con_apr_flatness.txt"
    fdat = np.loadtxt(fpath)
    fdat[10] = np.inf
    fdat[11] = -np.inf
    fdat[12] = 1.5
    fdat[13] = -1.4
    np.savetxt(fpath, fdat)

    samples, response, names = IndentationRater.load_training_set(
        path=tdir,
        which_type="continuous",
        replace_inf=True,  # This should be the default
        ret_names=True,
    )
    idf = names.index("feat_con_apr_flatness")
    data = samples[:, idf]

    assert np.allclose(data[10], 3)
    assert np.allclose(data[11], -3)
    assert np.allclose(data[12], 1.5)
    assert np.allclose(data[13], -1.4)
